from __future__ import annotations

import json
import os
import re
from datetime import datetime, timedelta, timezone
from pathlib import Path

from diracx.core.models import TokenResponse

EXPIRES_GRACE_SECONDS = 15


def dotenv_files_from_environment(prefix: str) -> list[str]:
    """Get the sorted list of .env files to use for configuration"""
    env_files = {}
    for key, value in os.environ.items():
        if match := re.fullmatch(rf"{prefix}(?:_(\d+))?", key):
            env_files[int(match.group(1) or -1)] = value
    return [v for _, v in sorted(env_files.items())]


def serialize_credentials(token_response: TokenResponse) -> str:
    """Serialize DiracX client credentials to a string

    This method is separated from write_credentials to allow for DIRAC to be
    able to serialize credentials for inclusion in the proxy file.
    """
    expires = datetime.now(tz=timezone.utc) + timedelta(
        seconds=token_response.expires_in - EXPIRES_GRACE_SECONDS
    )
    credential_data = {
        "access_token": token_response.access_token,
        "refresh_token": token_response.refresh_token,
        "expires_on": int(datetime.timestamp(expires)),
    }
    return json.dumps(credential_data)


def write_credentials(token_response: TokenResponse, *, location: Path | None = None):
    """Write credentials received in dirax_preferences.credentials_path"""
    from diracx.core.preferences import get_diracx_preferences

    credentials_path = location or get_diracx_preferences().credentials_path
    credentials_path.parent.mkdir(parents=True, exist_ok=True)
    credentials_path.write_text(serialize_credentials(token_response))



from __future__ import annotations

from diracx.core.utils import dotenv_files_from_environment


def test_dotenv_files_from_environment(monkeypatch):
    monkeypatch.setattr("os.environ", {})
    assert dotenv_files_from_environment("TEST_PREFIX") == []

    monkeypatch.setattr("os.environ", {"TEST_PREFIX": "/a"})
    assert dotenv_files_from_environment("TEST_PREFIX") == ["/a"]

    monkeypatch.setattr("os.environ", {"TEST_PREFIX": "/a", "TEST_PREFIX_1": "/b"})
    assert dotenv_files_from_environment("TEST_PREFIX") == ["/a", "/b"]

    monkeypatch.setattr(
        "os.environ",
        {"TEST_PREFIX_2": "/c", "TEST_PREFIX": "/a", "TEST_PREFIX_1": "/b"},
    )
    assert dotenv_files_from_environment("TEST_PREFIX") == ["/a", "/b", "/c"]

    monkeypatch.setattr(
        "os.environ",
        {"TEST_PREFIX_2a": "/c", "TEST_PREFIX": "/a", "TEST_PREFIX_1": "/b"},
    )
    assert dotenv_files_from_environment("TEST_PREFIX") == ["/a", "/b"]



from http import HTTPStatus


class DiracHttpResponse(RuntimeError):
    def __init__(self, status_code: int, data):
        self.status_code = status_code
        self.data = data


class DiracError(RuntimeError):
    http_status_code = HTTPStatus.BAD_REQUEST  # 400
    http_headers: dict[str, str] | None = None

    def __init__(self, detail: str = "Unknown"):
        self.detail = detail


class AuthorizationError(DiracError):
    ...


class PendingAuthorizationError(AuthorizationError):
    """Used to signal the device flow the authentication is still ongoing"""


class ExpiredFlowError(AuthorizationError):
    """Used only for the Device Flow when the polling is expired"""


class ConfigurationError(DiracError):
    """Used whenever we encounter a problem with the configuration"""


class BadConfigurationVersion(ConfigurationError):
    """The requested version is not known"""


class InvalidQueryError(DiracError):
    """It was not possible to build a valid database query from the given input"""


class JobNotFound(Exception):
    def __init__(self, job_id: int):
        self.job_id: int = job_id
        super().__init__(f"Job {job_id} not found")



from __future__ import annotations

__all__ = ("Config", "ConfigSource", "LocalGitConfigSource")

import logging
import os
from abc import ABCMeta, abstractmethod
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any

import git
import yaml
from cachetools import Cache, LRUCache, TTLCache, cachedmethod
from pydantic import AnyUrl, parse_obj_as

from ..exceptions import BadConfigurationVersion
from .schema import Config

if TYPE_CHECKING:
    from pydantic.config import BaseConfig
    from pydantic.fields import ModelField

DEFAULT_CONFIG_FILE = "default.yml"
DEFAULT_CS_CACHE_TTL = 5
MAX_CS_CACHED_VERSIONS = 1

logger = logging.getLogger(__name__)


class ConfigSourceUrl(AnyUrl):
    host_required = False

    @classmethod
    # TODO: This should return ConfigSourceUrl but pydantic's type hints are wrong
    def validate(cls, value: Any, field: ModelField, config: BaseConfig) -> AnyUrl:
        """Overrides AnyUrl.validate to add file:// scheme if not present."""
        if isinstance(value, str) and "://" not in value:
            value = f"git+file://{value}"
        return super().validate(value, field, config)


class ConfigSource(metaclass=ABCMeta):
    """
    This classe is the one
    """

    __registry: dict[str, type[ConfigSource]] = {}
    scheme: str

    @abstractmethod
    def __init__(self, *, backend_url: ConfigSourceUrl) -> None:
        ...

    @abstractmethod
    def latest_revision(self) -> tuple[str, datetime]:
        ...

    @abstractmethod
    def read_raw(self, hexsha: str, modified: datetime) -> Config:
        ...

    def __init_subclass__(cls) -> None:
        if cls.scheme in cls.__registry:
            raise TypeError(f"{cls.scheme=} is already define")
        cls.__registry[cls.scheme] = cls

    @classmethod
    def create(cls):
        return cls.create_from_url(backend_url=os.environ["DIRACX_CONFIG_BACKEND_URL"])

    @classmethod
    def create_from_url(
        cls, *, backend_url: ConfigSourceUrl | Path | str
    ) -> ConfigSource:
        url = parse_obj_as(ConfigSourceUrl, str(backend_url))
        return cls.__registry[url.scheme](backend_url=url)

    def read_config(self) -> Config:
        """
        :raises:
            git.exc.BadName if version does not exist
        """
        hexsha, modified = self.latest_revision()
        return self.read_raw(hexsha, modified)

    def clear_caches(self):  # noqa
        pass


class LocalGitConfigSource(ConfigSource):
    scheme = "git+file"

    def __init__(self, *, backend_url: ConfigSourceUrl) -> None:
        if not backend_url.path:
            raise ValueError("Empty path for LocalGitConfigSource")

        repo_location = Path(backend_url.path)
        self.repo_location = repo_location
        self.repo = git.Repo(repo_location)
        self._latest_revision_cache: Cache = TTLCache(
            MAX_CS_CACHED_VERSIONS, DEFAULT_CS_CACHE_TTL
        )
        self._read_raw_cache: Cache = LRUCache(MAX_CS_CACHED_VERSIONS)

    def __hash__(self):
        return hash(self.repo_location)

    def clear_caches(self):
        self._latest_revision_cache.clear()
        self._read_raw_cache.clear()

    @cachedmethod(lambda self: self._latest_revision_cache)
    def latest_revision(self) -> tuple[str, datetime]:
        try:
            rev = self.repo.rev_parse("master")
        except git.exc.ODBError as e:  # type: ignore
            raise BadConfigurationVersion(f"Error parsing latest revision: {e}") from e
        modified = rev.committed_datetime.astimezone(timezone.utc)
        logger.debug(
            "Latest revision for %s is %s with mtime %s", self, rev.hexsha, modified
        )
        return rev.hexsha, modified

    @cachedmethod(lambda self: self._read_raw_cache)
    def read_raw(self, hexsha: str, modified: datetime) -> Config:
        """
        Returns the raw data from the git repo

        :returns hexsha, commit time, data
        """
        logger.debug("Reading %s for %s with mtime %s", self, hexsha, modified)
        rev = self.repo.rev_parse(hexsha)
        blob = rev.tree / DEFAULT_CONFIG_FILE
        raw_obj = yaml.safe_load(blob.data_stream.read().decode())
        config = Config.parse_obj(raw_obj)
        config._hexsha = hexsha
        config._modified = modified
        return config



import pytest

from diracx.core.extensions import extensions_by_priority


def test_extensions_by_priority(monkeypatch):
    monkeypatch.setenv("DIRACX_EXTENSIONS", "diracx")
    assert list(extensions_by_priority()) == ["diracx"]

    monkeypatch.setenv("DIRACX_EXTENSIONS", "os")
    assert list(extensions_by_priority()) == ["os"]

    monkeypatch.setenv("DIRACX_EXTENSIONS", "os,diracx")
    assert list(extensions_by_priority()) == ["os", "diracx"]

    monkeypatch.setenv("DIRACX_EXTENSIONS", "diracx,os")
    assert list(extensions_by_priority()) == ["diracx", "os"]

    monkeypatch.delenv("DIRACX_EXTENSIONS")
    assert list(extensions_by_priority()) == ["diracx"]

    monkeypatch.setenv("DIRACX_EXTENSIONS", "missingdiracx")
    with pytest.raises(RuntimeError, match="Could not find extension module"):
        list(extensions_by_priority())



from __future__ import annotations

__all__ = (
    "SqlalchemyDsn",
    "LocalFileUrl",
    "ServiceSettingsBase",
)

import contextlib
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncIterator, Self, TypeVar

from authlib.jose import JsonWebKey
from pydantic import AnyUrl, BaseSettings, SecretStr, parse_obj_as

if TYPE_CHECKING:
    from pydantic.config import BaseConfig
    from pydantic.fields import ModelField


T = TypeVar("T")


class SqlalchemyDsn(AnyUrl):
    allowed_schemes = {"sqlite+aiosqlite", "mysql+aiomysql"}


class TokenSigningKey(SecretStr):
    jwk: JsonWebKey

    def __init__(self, data: str):
        super().__init__(data)
        self.jwk = JsonWebKey.import_key(self.get_secret_value())

    @classmethod
    # TODO: This should return TokenSigningKey but pydantic's type hints are wrong
    def validate(cls, value: Any) -> SecretStr:
        """Load private keys from files if needed"""
        if isinstance(value, str) and not value.strip().startswith("-----BEGIN"):
            url = parse_obj_as(LocalFileUrl, value)
            value = Path(url.path).read_text()
        return super().validate(value)


class LocalFileUrl(AnyUrl):
    host_required = False
    allowed_schemes = {"file"}

    @classmethod
    # TODO: This should return LocalFileUrl but pydantic's type hints are wrong
    def validate(cls, value: Any, field: ModelField, config: BaseConfig) -> AnyUrl:
        """Overrides AnyUrl.validate to add file:// scheme if not present."""
        if isinstance(value, str) and "://" not in value:
            value = f"file://{value}"
        return super().validate(value, field, config)


class ServiceSettingsBase(BaseSettings, allow_mutation=False):
    @classmethod
    def create(cls) -> Self:
        raise NotImplementedError("This should never be called")

    @contextlib.asynccontextmanager
    async def lifetime_function(self) -> AsyncIterator[None]:
        """A context manager that can be used to run code at startup and shutdown."""
        yield
